#include <exception>
#include <QLibrary>
#include <QDebug>
#include <QDir>
#include <QThread>

// Заголовочные файлы ДТО9
#include "idisplay.h"
#include "ifptr.h"
#include "dto_errors.h"
#include "dto_const.h"

static const bool USE_SHOW_PROPERTIES = false;
static const bool DISPLAY_ON_FPTR = false;

#if QT_VERSION < QT_VERSION_CHECK(5, 0, 0)
#  include <QTextCodec>

static class TrInit
{
public:
    TrInit()
    {
        QTextCodec::setCodecForTr(QTextCodec::codecForName("utf-8"));
        QTextCodec::setCodecForCStrings(QTextCodec::codecForName("utf-8"));
        QTextCodec::setCodecForLocale(QTextCodec::codecForName("utf-8"));
    }
} g_TrInit;
#endif

class Exception : public std::exception {
public:
    Exception(const QString &msg)
        : m_message(msg)
    {
    }

    virtual ~Exception() throw()
    {
    }

    virtual const char* what() const throw()
    {
        static QByteArray rawMessage = m_message.toUtf8();
        return rawMessage.data();
    }

private:
    QString m_message;
};

class Thread : public QThread
{
    Q_OBJECT
public:
    static void sleep(unsigned long secs)
    {
        QThread::sleep(secs);
    }
};

void checkError(TED::IDTOBase *iface)
{
    if(!iface)
        throw Exception("Invalid interface");

    int rc = EC_OK;
    iface->get_ResultCode(&rc);
    if(rc < 0)
    {
        QString resultDescription, badParamDescription;
        QVector<wchar_t> v(256);
        int size = iface->get_ResultDescription(&v[0], v.size());
        if (size <= 0)
            throw Exception("get_ResultDescription error");
        if (size > v.size())
        {
            v.clear();
            v.resize(size + 1);
            iface->get_ResultDescription(&v[0], v.size());
        }
        resultDescription = QString::fromWCharArray(&v[0]);
        if(rc == EC_INVALID_PARAM) {
            QVector<wchar_t> v(256);
            int size = iface->get_BadParamDescription(&v[0], v.size());
            if (size <= 0)
                throw Exception("get_BadParamDescription error");
            if (size > v.size())
            {
                v.clear();
                v.resize(size + 1);
                iface->get_ResultDescription(&v[0], v.size());
            }
            badParamDescription = QString::fromWCharArray(&v[0]);
        }

        if(badParamDescription.isEmpty())
            throw Exception(QObject::trUtf8("[%1] %2")
                            .arg(rc)
                            .arg(resultDescription));
        else
            throw Exception(QObject::trUtf8("[%1] %2 (%3)")
                            .arg(rc)
                            .arg(resultDescription)
                            .arg(badParamDescription));
    }
}

int main()
{
#if defined(Q_OS_LINUX)
    QLibrary libFptr("../bins/linux-x64/libfptr.so");
    if(!libFptr.load())
    {
        libFptr.setFileName("../bins/linux-x86/libfptr.so");
        libFptr.load();
    }
#elif defined(Q_OS_WIN32)
    QLibrary libFptr("../bins/nt-x86-mingw/fptr.dll");
    libFptr.load();
#else
#   error "Unknown OS"
#endif

    if(!libFptr.isLoaded())
    {
        qDebug() << qPrintable(QObject::trUtf8("Не удалось загрузить библиотеку fptr - [%1]")
                               .arg(libFptr.errorString()));
        return 1;
    }


#if defined(Q_OS_LINUX)
    QLibrary libDisplay("../bins/linux-x64/libdisplay.so");
    if(!libDisplay.load())
    {
        libDisplay.setFileName("../bins/linux-x86/libdisplay.so");
        libDisplay.load();
    }
#elif defined(Q_OS_WIN32)
    QLibrary libDisplay("../bins/nt-x86-mingw/display.dll");
    libDisplay.load();
#else
#   error "Unknown OS"
#endif

    if(!libDisplay.isLoaded())
    {
        qDebug() << qPrintable(QObject::trUtf8("Не удалось загрузить библиотеку display - [%1]")
                               .arg(libDisplay.errorString()));
        return 1;
    }

    CreateFptrInterfacePtr createFptr = (CreateFptrInterfacePtr) libFptr.resolve("CreateFptrInterface");
    ReleaseFptrInterfacePtr releaseFptr = (ReleaseFptrInterfacePtr) libFptr.resolve("ReleaseFptrInterface");
    CreateDisplayInterfacePtr createDisplay = (CreateDisplayInterfacePtr) libDisplay.resolve("CreateDisplayInterface");
    ReleaseDisplayInterfacePtr releaseDisplay = (ReleaseDisplayInterfacePtr) libDisplay.resolve("ReleaseDisplayInterface");

    TED::Fptr::IFptr *ifptr = createFptr(DTO_IFPTR_VER1);
    if(!ifptr)
    {
        qDebug() << qPrintable("Cannot create Fptr object");
        return 1;
    }
    TED::Display::IDisplay *idisplay = createDisplay(DTO_IDISPLAY_VER1);
    if(!idisplay)
    {
        qDebug() << qPrintable("Cannot create Display object");
        return 1;
    }

    try
    {
        // Выставляем рабочий каталог. В нем дККМ и дДП будут искать требуемые им библиотеки.
        QFileInfo finfo(libFptr.fileName());
        if(ifptr->put_DeviceSingleSetting(S_SEARCHDIR, finfo.absolutePath().toStdWString().c_str()) < 0)
            checkError(ifptr);
        if(ifptr->ApplySingleSettings() < 0)
            checkError(ifptr);
        if(idisplay->put_DeviceSingleSetting(S_SEARCHDIR, finfo.absolutePath().toStdWString().c_str()) < 0)
            checkError(idisplay);
        if(idisplay->ApplySingleSettings() < 0)
            checkError(idisplay);

        // Для настройки драйвера можно вызвать графическое окно настроек
        if(USE_SHOW_PROPERTIES)
        {
            if(ifptr->ShowProperties() < 0)
                checkError(ifptr);
            if(idisplay->ShowProperties() < 0)
                checkError(idisplay);
        }
        // Или настроить без него
        else
        {
            // COM17
            //if(ifptr->put_DeviceSingleSetting(S_PORT, 17) < 0)
                //checkError(ifptr);
            // USB. Можно указать положение на шине (USB$1-1.3, например), но тогда не нужно указывать Vid и Pid
            // Работа напрямую с USB - Linux Only!
            if(ifptr->put_DeviceSingleSetting(S_PORT, SV_PORT_USB) < 0)
                checkError(ifptr);
            if(ifptr->put_DeviceSingleSetting(S_VID, 0x2912) < 0)
                checkError(ifptr);
            if(ifptr->put_DeviceSingleSetting(S_PID, 0x0001) < 0)
                checkError(ifptr);
            if(ifptr->put_DeviceSingleSetting(S_MODEL, TED::Fptr::ModelFPrintPay01K) < 0)
                checkError(ifptr);
            if(ifptr->put_DeviceSingleSetting(S_PROTOCOL, TED::Fptr::ProtocolAtol30) < 0)
                checkError(ifptr);
            if(ifptr->put_DeviceSingleSetting(S_BAUDRATE, 115200) < 0)
                checkError(ifptr);
            if(ifptr->ApplySingleSettings() < 0)
                checkError(ifptr);

            // Для работы со стандратными ДП нужно настроить параметры связи с ними
            if(!DISPLAY_ON_FPTR)
            {
                if(idisplay->put_DeviceSingleSetting(S_PORT, SV_PORT_TTY) < 0)
                    checkError(idisplay);
                if(idisplay->put_DeviceSingleSetting(S_DEVICEFILE, L"ttyACM0") < 0)
                    checkError(idisplay);
                if(idisplay->put_DeviceSingleSetting(S_BAUDRATE, 115200) < 0)
                    checkError(idisplay);
            }
            // Для работы с дисплеем, подключенным к ККМ, нужно установить специальный порт.
            // Также согласуйте скорости портов в таблицах ККМ.
            else
            {
                if(idisplay->put_DeviceSingleSetting(S_PORT, SV_PORT_SOFT) < 0)
                    checkError(idisplay);
            }
            if(idisplay->put_DeviceSingleSetting(S_MODEL, TED::Display::ModelEpson) < 0)
                checkError(idisplay);
            if(idisplay->ApplySingleSettings() < 0)
                checkError(idisplay);

            // Для работы через ККМ необходимо связать дККМ и дДП
            if(DISPLAY_ON_FPTR)
            {
                void *hndl;
                ifptr->get_ClsPtr(&hndl);
                idisplay->put_ClsPtr(hndl);
            }

            // Активируем дККМ
            if (ifptr->put_DeviceEnabled(1) < 0)
                checkError(ifptr);
            // Активируем дДП
            if (idisplay->put_DeviceEnabled(1) < 0)
                checkError(idisplay);

            // Выводим текст
            idisplay->DisplayText(L"ТЕКСТ", TED::Display::ScrollLeft, 1000, 100);

            Thread::sleep(10);
        }
    }
    catch(const std::exception &ex)
    {
        qDebug() << qPrintable(ex.what());
        releaseFptr(&ifptr);
        releaseDisplay(&idisplay);
        return 1;
    }

    releaseFptr(&ifptr);
    releaseDisplay(&idisplay);
    return 0;
}

