#include <exception>
#include <QLibrary>
#include <QDebug>
#include <QDir>

// Заголовочные файлы ДТО9
#include "ifptr.h"
#include "ipaycard.h"
#include "dto_errors.h"
#include "dto_const.h"

#if QT_VERSION < QT_VERSION_CHECK(5, 0, 0)
#  include <QTextCodec>

static class TrInit
{
public:
    TrInit()
    {
        QTextCodec::setCodecForTr(QTextCodec::codecForName("utf-8"));
        QTextCodec::setCodecForCStrings(QTextCodec::codecForName("utf-8"));
        QTextCodec::setCodecForLocale(QTextCodec::codecForName("utf-8"));
    }
} g_TrInit;
#endif

class Exception : public std::exception {
public:
    Exception(const QString &msg)
        : m_message(msg)
    {
    }

    virtual ~Exception() throw()
    {
    }

    virtual const char* what() const throw()
    {
        static QByteArray rawMessage = m_message.toUtf8();
        return rawMessage.data();
    }

private:
    QString m_message;
};

void checkError(TED::IDTOBase *iface)
{
    if(!iface)
        throw Exception("Invalid interface");

    int rc = EC_OK;
    iface->get_ResultCode(&rc);
    if(rc < 0)
    {
        QString resultDescription, badParamDescription;
        QVector<wchar_t> v(256);
        int size = iface->get_ResultDescription(&v[0], v.size());
        if (size <= 0)
            throw Exception("get_ResultDescription error");
        if (size > v.size())
        {
            v.clear();
            v.resize(size + 1);
            iface->get_ResultDescription(&v[0], v.size());
        }
        resultDescription = QString::fromWCharArray(&v[0]);
        if(rc == EC_INVALID_PARAM) {
            QVector<wchar_t> v(256);
            int size = iface->get_BadParamDescription(&v[0], v.size());
            if (size <= 0)
                throw Exception("get_BadParamDescription error");
            if (size > v.size())
            {
                v.clear();
                v.resize(size + 1);
                iface->get_ResultDescription(&v[0], v.size());
            }
            badParamDescription = QString::fromWCharArray(&v[0]);
        }

        if(badParamDescription.isEmpty())
            throw Exception(QObject::trUtf8("[%1] %2")
                            .arg(rc)
                            .arg(resultDescription));
        else
            throw Exception(QObject::trUtf8("[%1] %2 (%3)")
                            .arg(rc)
                            .arg(resultDescription)
                            .arg(badParamDescription));
    }
}

int main()
{
#if defined(Q_OS_LINUX)
    QLibrary libFptr("../bins/linux-x64/libfptr.so");
    if(!libFptr.load())
    {
        libFptr.setFileName("../bins/linux-x86/libfptr.so");
        libFptr.load();
    }
#elif defined(Q_OS_WIN32)
    QLibrary libFptr("../bins/nt-x86-mingw/fptr.dll");
    libFptr.load();
#else
#   error "Unknown OS"
#endif

    if(!libFptr.isLoaded())
    {
        qDebug() << qPrintable(QObject::trUtf8("Не удалось загрузить библиотеку fptr - [%1]")
                               .arg(libFptr.errorString()));
        return 1;
    }


#if defined(Q_OS_LINUX)
    QLibrary libPaycard("../bins/linux-x64/libpaycard.so");
    if(!libPaycard.load())
    {
        libPaycard.setFileName("../bins/linux-x86/libpaycard.so");
        libPaycard.load();
    }
#elif defined(Q_OS_WIN32)
    QLibrary libPaycard("../bins/nt-x86-mingw/paycard.dll");
    libPaycard.load();
#else
#   error "Unknown OS"
#endif

    if(!libPaycard.isLoaded())
    {
        qDebug() << qPrintable(QObject::trUtf8("Не удалось загрузить библиотеку paycard - [%1]")
                               .arg(libPaycard.errorString()));
        return 1;
    }

    CreateFptrInterfacePtr createFptr = (CreateFptrInterfacePtr) libFptr.resolve("CreateFptrInterface");
    ReleaseFptrInterfacePtr releaseFptr = (ReleaseFptrInterfacePtr) libFptr.resolve("ReleaseFptrInterface");
    CreatePayCardInterfacePtr createPaycard = (CreatePayCardInterfacePtr) libPaycard.resolve("CreatePayCardInterface");
    ReleasePayCardInterfacePtr releasePaycard = (ReleasePayCardInterfacePtr) libPaycard.resolve("ReleasePayCardInterface");



    TED::Fptr::IFptr *ifptr = createFptr(DTO_IFPTR_VER1);
    if(!ifptr)
    {
        qDebug() << qPrintable("Cannot create Fptr object");
        return 1;
    }
    TED::PayCard::IPayCard *ipaycard = createPaycard(DTO_IPAYCARD_VER1);
    if(!ipaycard)
    {
        qDebug() << qPrintable("Cannot create PayCard object");
        return 1;
    }

    try
    {
        // Выставляем рабочий каталог. В нем дККМ и дПС будут искать требуемые им библиотеки.
        QFileInfo finfo(libFptr.fileName());
        if(ifptr->put_DeviceSingleSetting(S_SEARCHDIR, finfo.absolutePath().toStdWString().c_str()) < 0)
            checkError(ifptr);
        if(ifptr->ApplySingleSettings() < 0)
            checkError(ifptr);
        if(ipaycard->put_DeviceSingleSetting(S_SEARCHDIR, finfo.absolutePath().toStdWString().c_str()) < 0)
            checkError(ipaycard);
        if(ipaycard->ApplySingleSettings() < 0)
            checkError(ipaycard);

        // Для настройки драйвера можно вызвать графическое окно настроек
        if(false)
        {
            if(ifptr->ShowProperties() < 0)
                checkError(ifptr);
            if(ipaycard->ShowProperties() < 0)
                checkError(ipaycard);
        }
        // Или настроить без него
        else
        {
            // COM17
            //if(ifptr->put_DeviceSingleSetting(S_PORT, 17) < 0)
                //checkError(ifptr);
            // USB. Можно указать положение на шине (USB$1-1.3, например), но тогда не нужно указывать Vid и Pid
            // Работа напрямую с USB - Linux Only!
            if(ifptr->put_DeviceSingleSetting(S_PORT, SV_PORT_USB) < 0)
                checkError(ifptr);
            if(ifptr->put_DeviceSingleSetting(S_VID, 0x2912) < 0)
                checkError(ifptr);
            if(ifptr->put_DeviceSingleSetting(S_PID, 0x0001) < 0)
                checkError(ifptr);
            if(ifptr->put_DeviceSingleSetting(S_MODEL, TED::Fptr::ModelFPrintPay01K) < 0)
                checkError(ifptr);
            if(ifptr->put_DeviceSingleSetting(S_PROTOCOL, TED::Fptr::ProtocolAtol30) < 0)
                checkError(ifptr);
            if(ifptr->put_DeviceSingleSetting(S_BAUDRATE, 115200) < 0)
                checkError(ifptr);
            if(ifptr->ApplySingleSettings() < 0)
                checkError(ifptr);

            // Для работы с обычными пинпадами (USB, COM, TCP) нужно настроить параметры связи
            if (false)
            {
                if(ipaycard->put_DeviceSingleSetting(S_PROTOCOL, TED::PayCard::PaySystemArcomARCUS2) < 0)
                    checkError(ipaycard);
                if(ipaycard->put_DeviceSingleSetting(S_PORT, SV_PORT_USB) < 0)
                    checkError(ipaycard);
                if(ipaycard->put_DeviceSingleSetting(S_VID, 0x079B) < 0)
                    checkError(ipaycard);
                if(ipaycard->put_DeviceSingleSetting(S_PID, 0x0028) < 0)
                    checkError(ipaycard);
            }
            // Для работы с пинпадом FPrintPay выделены отдельные модели, им настройки связи не требуются
            else
            {
                if(ipaycard->put_DeviceSingleSetting(S_PROTOCOL, TED::PayCard::PaySystemArcomARCUS2FPrintPay) < 0)
                    checkError(ipaycard);
                // Настроим способ соединения с сервером авторизации платежной системы (в данном случае через сеть ПК).
                if(ipaycard->put_DeviceSingleSetting(S_MODEMMODE, TED::PayCard::ModemModeTcp) < 0)
                    checkError(ipaycard);
            }
            // Устанавливаем рабочий каталог платежной системы. В нем библиотека ПС будет хранить нужные ей в работе данные.
            // В нем должна лежать и сама библиотека.
            if(ipaycard->put_DeviceSingleSetting(S_PATHAC, finfo.absolutePath().toStdWString().c_str()) < 0)
                checkError(ipaycard);

            // Для разных платежных систем могут потребоваться дополнительные настройки.
            // Например, для инпаса нужен ID пинпада
            if(ipaycard->put_DeviceSingleSetting(S_TERMINALID, L"1234567890") < 0)
                checkError(ipaycard);

            if(ipaycard->ApplySingleSettings() < 0)
                checkError(ipaycard);

            // Для работы с FPrintPay нужно связать экземпляры дККМ и дПС,
            // чтобы дПС мог использовать дККМ как транспортный уровень обмена с пинпадом и модемом.
            if (true)
            {
                void *hndl;
                ifptr->get_PinPadDevice(&hndl);
                ipaycard->put_PinPadDevice(hndl);
                ifptr->get_ModemDevice(&hndl);
                ipaycard->put_ModemDevice(hndl);
            }

            // Процедура сохранения актуальных настроек дПС. Если пользуетесь ShowProperties(),
            // то ей можно вытащить актуальные настройки, чтобы потом передать из в драйвер через put_DeviceSettings(),
            // а не настраивать заново
            if(true)
            {
                QVector<wchar_t> v(256);
                int size = ipaycard->get_DeviceSettings(&v[0], v.size());
                if (size <= 0)
                    throw Exception("get_DeviceSettings error");
                if (size > v.size())
                {
                    v.clear();
                    v.resize(size + 1);
                    ipaycard->get_DeviceSettings(&v[0], v.size());
                }
                QString settings = QString::fromWCharArray(&v[0], v.size());
                // settings - строка с настройками
                qDebug() << qPrintable(settings);
            }

            // Активируем дККМ
            if (ifptr->put_DeviceEnabled(1) < 0)
                checkError(ifptr);
            if (ifptr->GetStatus() < 0)
                checkError(ifptr);
            int receiptWidth = 0;
            // Запоминаем кол-во символов в строке в ККМ
            ifptr->get_CharLineLength(&receiptWidth);
            // Активируем дПС
            if (ipaycard->put_DeviceEnabled(1) < 0)
                checkError(ipaycard);

            // Подготовка авторизации (оплата). Если нужно сделать возврат/отмену,
            // то нужно поставить соответствующую операцию и выставить параметры оригинальной транзакции.
            if (ipaycard->put_AuthorizationType(TED::PayCard::AuthorizationTypeReaderPIN) < 0)
                checkError(ipaycard);
            if (ipaycard->put_OperationType(TED::PayCard::OperationTypeSub) < 0)
                checkError(ipaycard);
            if (ipaycard->put_Sum(10.0) < 0)
                checkError(ipaycard);
            // Ставим ширину слипа в соответствии с шириной ленты ККМ
            if (ipaycard->put_CharLineLength(receiptWidth) < 0)
                checkError(ipaycard);

            // Далее идет запрос дополнительных параметров для ПС. Все их можно передать заранее, либо в этом месте выводить запросы кассиру (актуально для CyberPlat).
            while(ipaycard->PrepareAuthorization())
            {
                int rc = EC_OK;
                ipaycard->get_ResultCode(&rc);
                switch(rc)
                {
                case EC_PAYCARD_NEED_ACCOUNT:
                    break;
                case EC_PAYCARD_NEED_AUTH_CODE:
                    break;
                case EC_PAYCARD_NEED_PHONE_NUMBER:
                    break;
                case EC_PAYCARD_NEED_REFERENCE_NUMBER:
                    break;
                }
            }

            // Проводим транзакцию
            if (ipaycard->Authorization() < 0)
                checkError(ipaycard);

            // При удачно проведенной транзакции нужно сохранить куда-нибудь ее параметры,
            // чтобы потом можно было ее отменить и вернуть деньги на счет клиента.
            // Список возвращаемых параметров представлен в документации, можно его сохранить весь.

            // Печатаем слип как нефискальный документ
            QVector<wchar_t> v(2048);
            int size = ipaycard->get_Text(&v[0], v.size());
            if (size <= 0)
                throw Exception("get_DeviceSettings error");
            if (size > v.size())
            {
                v.clear();
                v.resize(size + 1);
                ipaycard->get_Text(&v[0], v.size());
            }
            QString slip = QString::fromWCharArray(&v[0], v.size());

            if(ifptr->put_Caption(slip.toStdWString().c_str()) < 0)
                checkError(ifptr);
            if(ifptr->put_TextWrap(TED::Fptr::TextWrapLine) < 0)
                checkError(ifptr);
            if(ifptr->put_Alignment(TED::Fptr::AlignmentLeft) < 0)
                checkError(ifptr);
            if(ifptr->PrintString() < 0)
                checkError(ifptr);

            if(ifptr->put_Mode(TED::Fptr::ModeReportNoClear) < 0)
                checkError(ifptr);
            if(ifptr->SetMode() < 0)
                checkError(ifptr);
            if(ifptr->PrintFooter() < 0)
                checkError(ifptr);


            // Закрытие банковской смены
            if (ipaycard->put_ReportType(TED::PayCard::ReportTypeCloseSession) < 0)
                checkError(ipaycard);
            if (ipaycard->put_CharLineLength(receiptWidth) < 0)
                checkError(ipaycard);
            if(ipaycard->BeginReport() < 0)
                checkError(ipaycard);
            if(ipaycard->EndReport() < 0)
                checkError(ipaycard);

            // Печатаем слип как нефискальный документ
            v.resize(2048);
            size = ipaycard->get_Text(&v[0], v.size());
            if (size <= 0)
                throw Exception("get_DeviceSettings error");
            if (size > v.size())
            {
                v.clear();
                v.resize(size + 1);
                ipaycard->get_Text(&v[0], v.size());
            }
            slip = QString::fromWCharArray(&v[0], v.size());

            if(ifptr->put_Caption(slip.toStdWString().c_str()) < 0)
                checkError(ifptr);
            if(ifptr->put_TextWrap(TED::Fptr::TextWrapLine) < 0)
                checkError(ifptr);
            if(ifptr->put_Alignment(TED::Fptr::AlignmentLeft) < 0)
                checkError(ifptr);
            if(ifptr->PrintString() < 0)
                checkError(ifptr);

            if(ifptr->put_Mode(TED::Fptr::ModeReportNoClear) < 0)
                checkError(ifptr);
            if(ifptr->SetMode() < 0)
                checkError(ifptr);
            if(ifptr->PrintFooter() < 0)
                checkError(ifptr);
        }
    }
    catch(const std::exception &ex)
    {
        qDebug() << qPrintable(ex.what());
        releaseFptr(&ifptr);
        releasePaycard(&ipaycard);
        return 1;
    }

    releaseFptr(&ifptr);
    releasePaycard(&ipaycard);
    return 0;
}

